use crate::read;
use std::collections::HashSet;
use itertools::Itertools;

#[derive(Debug, PartialEq, Eq, Default, Clone, Hash)]
struct Position {
    pub x: i32,
    pub y: i32,
}

impl Position {
    pub fn move_right(&self) -> Self {
        Self {
            x: self.x + 1,
            y: self.y,
        }
    }
    pub fn move_left(&self) -> Self {
        Self {
            x: self.x - 1,
            y: self.y,
        }
    }
    pub fn move_up(&self) -> Self {
        Self {
            x: self.x,
            y: self.y - 1,
        }
    }
    pub fn move_down(&self) -> Self {
        Self {
            x: self.x,
            y: self.y + 1,
        }
    }

    pub fn follow_head(&self, head: &Position) -> Self {
        // Based on jenarvaezg/aoc2022
        // https://old.reddit.com/r/adventofcode/comments/zgnice/2022_day_9_solutions/izi6s4p/
        if self.x.abs_diff(head.x) < 2 && self.y.abs_diff(head.y) < 2 {
            self.clone()
        } else {
            Self {
                x: self.x + (head.x - self.x).signum(),
                y: self.x + (head.y - self.y).signum(),
            }
        }
    }
}

struct Chain {
    pub knots: Vec<Position>,
    visited: HashSet<Position>
}

impl Chain {
    pub fn new(num_knots: usize) -> Self {
        Self {
            knots: vec![Position::default(); num_knots],
            visited: HashSet::with_capacity(10000)
        }
    }
    pub fn num_visited(&self) -> usize {
        self.visited.len()
    }

    pub fn move_knots(&mut self, direction: &str, count: usize) {
        let move_head_func = match direction {
            "R" => Position::move_right,
            "U" => Position::move_up,
            "L" => Position::move_left,
            "D" => Position::move_down,
            _ => unreachable!(),
        };
        for _c in 0..count {
            self.knots[0] = move_head_func(&mut self.knots[0]);
            for i in 1..self.knots.len() {
                self.knots[i] = self.knots[i].follow_head(&self.knots[i - 1]);
            }

            self.visited.insert(self.knots.last().unwrap().clone());
        }
    }
}

pub fn run() {
    let res = simulate(vec![2, 10]);
    #[cfg(feature = "part1")]
    {
        println!("Day 9, Part 01: {}", res[0]);
    }

    #[cfg(feature = "part2")]
    {
        println!("Day 9, Part 02: {}", res[1]);
    }
}

fn simulate(chain_sizes: Vec<usize>) -> Vec<usize> {
    let chains = read("09")
        .lines()
        .filter(|l|!l.is_empty())
        .map(|l|l.split_once(' ').unwrap())
        .map(|(dir, count)|(dir, count.parse::<usize>().unwrap()))
        .fold(chain_sizes
                  .iter()
                  .map(|c|Chain::new(*c))
                  .collect_vec(),
              |mut chains, (dir, count)| {

            chains
                .iter_mut()
                .for_each(|chain|chain.move_knots(dir, count));
            chains
        });
    chains.iter().map(|c|c.num_visited()).collect()
}
